#pylint: disable=C0103,W0201

"""
This file contains helper functions and classes that both the client and
server need to use.
"""

from Crypto.Random import random
from Crypto.Cipher import PKCS1_OAEP
from Crypto.Signature import PKCS1_v1_5
from Crypto.Hash import SHA
import string
import base64
import json
import time
import zlib

def encrypt_and_sign_dict(data, sender_private_key, recipient_public_key):
    """
    Uniquify the dict (by adding a timestamp), sign and encrypt it with the
    provided keys, and then return the result. This also takes care of
    encoding everything correctly. Additionally, we zlib compress the data
    because we're often trying to encrypt data that's quite large -- almost
    the size of the RSA key used to encrypt it -- and RSA doesn't let us
    encrypt data larger than our key size.
    """

    data['timestamp'] = time.time()
    plaintext = zlib.compress(json.dumps(data))

    # PyCrypto's docs for RSA recommend this encryption algorithm over RSA
    cipher = PKCS1_OAEP.new(recipient_public_key)
    ciphertext = cipher.encrypt(plaintext)

    # PyCrypto's docs for RSA recommend this signing algorithm over RSA
    h = SHA.new(ciphertext)
    signer = PKCS1_v1_5.new(sender_private_key)
    signature = signer.sign(h)

    # base64 encode so that we can uniquely split on the colon
    return base64.b64encode(ciphertext) + ":" + base64.b64encode(signature)

def verify_and_decrypt_dict(message, recipient_private_key, sender_public_key):
    """
    Verify and decrypt a message with the provided keys, and then return the
    result. This also takes care of decompressing and decoding everything
    correctly. If anything goes wrong, throw an exception.
    """

    ciphertext, signature = message.split(":")
    ciphertext = base64.b64decode(ciphertext)
    signature = base64.b64decode(signature)

    h = SHA.new(ciphertext)
    verifier = PKCS1_v1_5.new(sender_public_key)
    if not verifier.verify(h, signature):
        raise Exception

    cipher = PKCS1_OAEP.new(recipient_private_key)
    plaintext = cipher.decrypt(ciphertext)

    return json.loads(zlib.decompress(plaintext))

def f(pk, x):
    """The f function defined by Damgard and Nielsen"""
    return pow(pk.g, x, pk.p)

def verify_in_image(pk, x):
    """
    Given a public key, verify that x in image(f), otherwise, throw an
    exception.
    """
    if x < 1 or x >= pk.p or pow(x, pk.q, pk.p) != 1:
        raise Exception

class PublicKey(object):
    """The public key used by the commitment scheme"""

    @staticmethod
    def from_dsa(dsa_key, private_key):
        """Create a key from a given DSA key and private key value"""

        self = PublicKey()
        self.p = dsa_key.p
        self.q = dsa_key.q
        self.g = dsa_key.g
        self.y = f(self, private_key)
        return self

    def modify_private_key(self, f_private_key=None):
        """
        Modify our private key using a partial private key from the other
        party. See our paper for more details on why we can't have a single
        party pick the private key.
        """

        private_key = None
        if f_private_key is None:
            private_key = random.randint(1, self.q-1)
            f_private_key = f(self, private_key)

        self.y = (self.y * f_private_key) % self.p
        self.check_y()
        return private_key, f_private_key

    def random_group_value(self):
        """Return a random value in the domain of f"""

        return random.randint(0, self.q-1)

    def to_dict(self):
        """Convert the key to a dict"""

        return dict(
            p = self.p,
            q = self.q,
            g = self.g,
            y = self.y,
        )

    @staticmethod
    def from_dict(data):
        """Create a key object from a dict returned by to_dict"""

        self = PublicKey()
        self.p = data['p']
        self.q = data['q']
        self.g = data['g']
        self.y = data['y']
        self.check_y()
        return self

    def check_y(self):
        """The commitment scheme fails if y==1 or y notin image(f)"""
        verify_in_image(self, self.y)
        if self.y == 1:
            raise Exception

class CommittedValue(object):
    """
    This class represents a value, its commitment, and the knowledge needed to
    reveal the commmitment. (Compare with the Commitment class, which
    represents just a commitment without knowledge of the value it
    represents.)
    """

    def __init__(self, val, rnd, pk):
        self.val = val
        if self.val is None:
            self.val = pk.random_group_value()

        self.rnd = rnd
        if self.rnd is None:
            self.rnd = pk.random_group_value()

        self.cmt = (pow(pk.y, self.val, pk.p) * f(pk, self.rnd)) % pk.p

    @staticmethod
    def add(c1, c2, pk):
        """Add two CommitedValues to produce a new one"""

        new_val = (c1.val + c2.val) % pk.q
        new_rnd = c1.rnd + c2.rnd
        return CommittedValue(new_val, new_rnd, pk)

    @staticmethod
    def invert(c, pk):
        """Return the inverse of a CommitedValue"""

        new_val = (-c.val) % pk.q
        new_rnd = (-c.rnd) % pk.q
        return CommittedValue(new_val, new_rnd, pk)

    @staticmethod
    def randomize(c, pk):
        """Return a new CommitedValue with the same value but new randomness"""

        # generate new randomness
        return CommittedValue(c.val, None, pk)

    def get_commitment(self):
        """Return just the commitment part of this object"""
        return self.cmt

    def reveal(self):
        """Return a dict of info needed to reveal this commitment"""

        return dict(
            val = self.val,
            rnd = self.rnd,
            cmt = self.cmt,
        )

class Commitment(object):
    """This class represents a commitment whose value is unknown"""

    def __init__(self, cmt, pk):
        # As per page 22 of http://www.brics.dk/RS/97/27/BRICS-RS-97-27.pdf,
        # whenever we receive a commitment, we must verify that the commitment
        # is in the image of f
        verify_in_image(pk, cmt)
        self.cmt = cmt

    def get_commitment(self):
        """Return the commitment"""
        return self.cmt

    def check_and_get_committed_value(self, data, pk):
        """
        Given a dict (from the reveal method of a CommittedValue), first
        verify that the info provided is a valid reveal for this commitment.
        If it is, return a CommittedValue with the value. If not, raise an
        exception.
        """

        val = data['val']
        rnd = data['rnd']
        reference = CommittedValue(val, rnd, pk)
        if reference.cmt != self.cmt or reference.cmt != data['cmt']:
            raise Exception
        return reference

    @staticmethod
    def add(c1, c2, pk):
        """Add two Commitments to produce a new one"""

        new_cmt = (c1.cmt * c2.cmt) % pk.p
        return Commitment(new_cmt, pk)

    @staticmethod
    def invert(c, pk):
        """Return the inverse of a Commitment"""

        new_cmt = pow(c.cmt, pk.q-1, pk.p)
        return Commitment(new_cmt, pk)

class Password(object):
    """
    This class stores passwords, and in particular, interfaces between the two
    representations we use. Usually, passwords are thought of as integers.
    However, when a password is displayed to a user, it is displayed as a
    string. In order to make passwords easy to uniquely read/type, we avoid
    potentially ambiguous/look-alike characters. Of course, we can uniquely
    convert between the two representations.
    """

    # use all uppercase letters, lowercase letters, and digits, except for 0O1Il
    ALPHABET = string.ascii_uppercase+string.ascii_lowercase+string.digits
    ALPHABET = ALPHABET.translate(None, "0O1Il")

    def __init__(self, num=None, txt=None):
        """
        Specify exactly one of the num or txt representations to create a
        Password
        """

        if num is not None:
            self.num = num
            self.txt = Password.num_to_txt(num)
        elif txt is not None:
            self.txt = txt
            self.num = Password.txt_to_num(txt)
        else:
            raise Exception

    @staticmethod
    def num_to_txt(x):
        """Convert from an integer to a string"""

        if x == 0:
            return Password.ALPHABET[0]

        # this is basically a conversion from base 10 to base 57, except we
        # don't care that the end result has its digits in the reverse order
        base = len(Password.ALPHABET)
        password = ""
        while x > 0:
            password = password + Password.ALPHABET[x % base]
            x = x / base
        return password

    @staticmethod
    def txt_to_num(s):
        """Convert from a string to an integer"""

        if s[-1] == Password.ALPHABET[0]:
            raise Exception
        x = 0
        i = 1
        base = len(Password.ALPHABET)
        for c in s:
            x += Password.ALPHABET.index(c) * i
            i = i * base
        return x
