#pylint: disable=C0301,F0401,C0103,R0902,W0201

"""
This is the server-side part of the password generation protocol.
"""

from Crypto.Random import random
from Crypto import Random
from Crypto.PublicKey import _DSA

# because the client and server are running on the same computer (right now),
# the lib directory is located elsewhere, and we have to hack $PATH to find it
import sys
sys.path.append('../lib')
from pwgen_lib import PublicKey, Commitment, CommittedValue, encrypt_and_sign_dict

import service_keys
import threading

def generate_dsa_key():
    """
    Create and return a new DSA key. Note that we have to use _DSA.generate_py
    rather than DSA.generate because the latter only lets us create keys of
    <=1024 bits, whereas the former is a helper function that bypasses this
    check.
    """
    return _DSA.generate_py(2048, Random.new().read)

spare_dsa_keys = []

class KeyGenerator(threading.Thread):
    """For creating DSA keys in the background"""

    def run(self):
        dsa_key = generate_dsa_key()
        spare_dsa_keys.append(dsa_key)

# The slowest phase of password generation is generating a unique DSA key per-
# user. To speed up this part of the process, we're going to attempt to always
# have an already-created-but-unused key waiting around. As soon as this
# module is imported, let's create one (and then every time we use a key, we
# start a background process to create a new one).
KeyGenerator().start()

class ServerPasswordGenerator(object):
    """Store everything in a wrapper class for convenience"""

    def __init__(self, username):
        self.username = username

        try:
            # if we already have an unused key generated, used that
            dsa_key = spare_dsa_keys.pop()
        except IndexError:
            dsa_key = generate_dsa_key()
        self.private_key = random.randint(1, dsa_key.q-1)
        self.public_key = PublicKey.from_dsa(dsa_key, self.private_key)

        # generate a new key in the background
        KeyGenerator().start()

    def get_public_key(self):
        """Return the public key for the commitment scheme, as a dict"""

        return self.public_key.to_dict()

    def modify_private_key(self, f_client_private_key):
        """
        Modify our private key using a partial private key from the other
        party. See our paper for more details on why we can't have a single
        party pick the private key.
        """

        self.public_key.modify_private_key(f_client_private_key)

    def generate_commitment_s(self):
        """
        Pick a random value S, generate a CommitedValue for it, and return the
        associated commitment.
        """

        self.committed_S = CommittedValue(None, None, self.public_key)
        self.commitment_S = Commitment(self.committed_S.cmt, self.public_key)
        return self.committed_S.get_commitment()

    def store_commitment_c(self, cmt):
        """Store the commitment for C, sent by the client"""

        self.commitment_C = Commitment(cmt, self.public_key)

    def reveal_s(self):
        """Return the info needed to reveal S to the client"""

        return self.committed_S.reveal()

    def store_commitment_p(self, cmt):
        """Store the commitment for P, sent by the client"""

        self.commitment_P = Commitment(cmt, self.public_key)

    def verify_commitment_zero(self, cmt):
        """
        We just received a commitment from the client that claims to be for
        C+S-P. We indepdently compute what this commitment should be and
        verify that the two match, raising an exception otherwise. (Despite
        the name of this method, we do *not* check that this commitment is for
        the value of 0.)
        """

        commitment_sum = Commitment.add(self.commitment_S, self.commitment_C, self.public_key)
        inverse_P = Commitment.invert(self.commitment_P, self.public_key)
        self.commitment_computed_zero = Commitment.add(commitment_sum, inverse_P, self.public_key)
        if self.commitment_computed_zero.cmt != cmt:
            raise Exception

    def verify_reveal_zero(self, data):
        """
        Now, we take the commitment for C+S-P (which, thanks to
        verify_commitment_zero, we are guaranteed is in fact a commitment for
        C+S-P) and confirm that the client has revealed to us that this
        commitment is for the value of 0. This ensures that C+S=P.
        """

        committed_zero = self.commitment_computed_zero.check_and_get_committed_value(data, self.public_key)
        if committed_zero.val != 0:
            raise Exception

    def get_approval_messages(self, private_key):
        """
        Prepare a message (to be sent to the authentication server) approving
        of this commitment_P (with associated public_key) for this user.
        Unfortunately, HTTP doesn't authenticate sender, so we must sign and
        encrypt the message to prove it came from the the correct service. It
        turns out that this message is too big to encrypt, so we split it into
        two parts: the first part is signed and encrypted, and causes the two
        services to open a cookiejar. This cookie "authenticates" the sender
        of the second message, which is sent unencrypted. It's true that
        someone could MITM and tamper with this second message, but (a) that
        problem also exists elsewhere in this proof-of-concept website and
        we've chosen to ignore it, (b) even if an attacker modifies the
        contents of the second message, all he can do is prevent a legitimate
        user from registering -- but he can't impersonate or log in as anyone.
        This method returns, but does not send, the messages, and (c) as
        stated in known_bugs.txt, when we switch from HTTP to socket-based
        inter-service communication, this entire hack becomes unnecessary
        """

        data = dict(
            username = self.username,
            commitment_P = self.commitment_P.cmt,
        )
        m1 = encrypt_and_sign_dict(data, private_key, service_keys.AUTH_SVC_PUBLIC_KEY)
        m2 = dict(
            username = self.username,
            public_key = self.get_public_key(),
        )
        return (m1, m2)
